/************** decoder.c
		Decoder/character set converter routines for
		listarchiver and nlist
		Written by Gabor Paller
					*************/
					
#include	<stdio.h>
#include	<string.h>
#include	"decoder.h"
#include	"config.h"
#define	N_EXTERN	extern
#include	"commonvars.h"

static char base64chars[] = "ABCDEFGHIJKLMNOPQRSTUVWXYZabcdefghijklmnopqrstuvwxyz0123456789+/=";

#define CHSTRUCT struct charsets
CHSTRUCT {
  char	csname[16];
  int	cscode;
};

CHSTRUCT charset_table[] = {
  { "iso-8859-1",   CHARSET_ISO_8859_1 },
  { "iso-8859-2",   CHARSET_ISO_8859_2 },
  { "iso-8859_1",   CHARSET_ISO_8859_1 },
  { "iso-8859_2",   CHARSET_ISO_8859_2 },
  { "Windows-1250", CHARSET_WIN_1250 },
  { "Windows-1252", CHARSET_WIN_1252 },
  { "us-ascii",	    CHARSET_US_ASCII },
  { "utf-8",	    CHARSET_UTF_8 },
  { "x-user-defined",	CHARSET_USER_DEFINED },
  { "", 0 } 
};

/* Decodes a base64-coded line and returns the resulting line length */
int decode_base64( char *buffer, unsigned char *dest ) {
  unsigned char db[3];
  unsigned char *dp;
  char *sp,*p;
  int len = 0;
  int bp = 0,index;
  
  sp = buffer;
  dp = dest;
  while( *buffer && ( *buffer != '\n' ) ) {
 /* Don't accept non-base64 characters */
    if( ( p = strchr( base64chars,(int)*buffer ) ) == NULL )
    	return 0;
    index = p - base64chars;
    if( index >= 64 )
    	break;
    switch( bp ) {
      case 0:		/* Top of db[0] */
        db[1] = 0;
        db[2] = 0;
      	db[0] = (unsigned char)( index << 2 );
      	break;
      	
      case 1:		/* 2 LSBs of db[0], 4 MSBs of db[1] */
      	db[0] |= (unsigned char)( ( index & 0x30 ) >> 4 );
      	db[1] |= (unsigned char)( ( index & 0x0F ) << 4 );
      	break;
      	
      case 2:		/* 4 LSBs of db[1], 2 MSBs of db[2] */
      	db[1] |= (unsigned char)( ( index & 0x3C ) >> 2 );
      	db[2] |= (unsigned char)( ( index & 0x03 ) << 6 );
      	break;
      	
      case 3:
      	db[2] |= (unsigned char)( index & 0x3F );
      	break; 
    }
    ++bp;
    if( bp > 3 ) {
      *(dp++) = db[0];
      *(dp++) = db[1];
      *(dp++) = db[2];
      bp = 0;
      len += 3;     
    }
    ++buffer;
  }
  switch( bp ) {    	
    case 2:
    	*(dp++) = db[0];
    	++len;
    	break;
    	
    case 3:
    	*(dp++) = db[0];
    	*(dp++) = db[1];
    	len += 2;
    	break;
  }
  return len;
}

/* Decodes a quoted-printable line and returns the resulting line length */
int decode_quoted_printable( char *buffer, char *dest ) {
  char *dp,cv[3];
  int c,len = 0;
  
  dp = dest;
  while( *buffer ) {
    if( *buffer == '_' ) {
      *(dp++) = ' ';
      ++len;
      ++buffer;
    } else
    if( *buffer != '=' ) {
      *(dp++) = *(buffer++);
      ++len;
    } else
    if( ( *(buffer+1) !=  0 ) && ( *(buffer+2) != 0 ) ) {
      cv[0] = *(++buffer);
      cv[1] = *(++buffer);
      cv[2] = 0;
      ++buffer;
      if( sscanf( cv,"%x",&c ) < 1 )
      	continue;
      *(dp++) = (unsigned char)c;
      ++len;
    } else
    	++buffer;
  }
  *dp = 0;
  return len;
}

/* Decodes a string containing quoted substrings. Returns the string by
   overwriting its input buffer and the character set in the return value */
int decode_quoted_fragments( char buffer[] ) {
  char *sp,*dp,*fp,*p;
  int	charset = NO_CHARSET;
  int	encoding = ENCODING_NONE;
  int	len;
  CHSTRUCT	*csp;
  
  sp = dp = buffer;
  while( *sp ) {
/* Let's see if it is a quoted fragment starting with character set declaration
   then the encoding declaration then the quoted frgament itself */
    if( ( *sp == '=' ) && ( *(sp+1) == '?' ) ) {
      charset = NO_CHARSET;
      fp = sp;
      sp += 2;
      csp = charset_table;
      while( *( csp->csname ) ) {
        len = strlen( csp->csname );
        if( strncasecmp( sp,csp->csname,len ) == 0 ) {
          charset = csp->cscode;
          sp += len;
          break;
        }
        ++csp;
      }
      if( charset == NO_CHARSET ) {
illegal_char:
        sp = fp;
        *(dp++) = *(sp++);
      } else {
        if( *(sp++) != '?' )
        	goto illegal_char;
        switch( *(sp++) ) {
          case 'Q':
          case 'q':
          	encoding = ENCODING_QP;
          	break;
          
          case 'B':
          case 'b':
          	encoding = ENCODING_BASE64;
          	break;
          	
          default:
          	goto illegal_char;
        }
        if( *(sp++) != '?' )
        	goto illegal_char;
        if( ( p = strstr( sp,"?=" ) ) == NULL )
        	goto illegal_char;
        *p = 0;
        switch( encoding ) {
          case ENCODING_QP:
          	len = decode_quoted_printable( sp,dp );
          	break;
          	
          case ENCODING_BASE64:
          	len = decode_base64( sp,(unsigned char *)dp );
          	break;
        }
        dp += len;
        sp = p+2;
      }
    } else
    	*(dp++) = *(sp++);
  }
  *dp = 0;
  return charset;
}

/* Encodes a string according to the specified encoding to quoted printable format. */
void encode_quoted_printable( char buffer[], char dest[], int charset ) {
  CHSTRUCT	*csp;
  char		c,*d;
  bool		charset_found = FALSE;
  
  if( charset == CHARSET_US_ASCII ) {
return_original:
    strcpy( dest,buffer );
    return;
  }
  strcpy( dest,"=?" );
  csp = charset_table;
  while( *( csp->csname ) ) {
    if( csp->cscode == charset ) {
      strcat( dest,csp->csname );
      strcat( dest,"?Q?" );
      charset_found = TRUE;
      break;
    }
    ++csp;
  }
  if( !charset_found )
  	goto return_original;
  d = dest + strlen( dest );
  while( c = *(buffer++) ) {
    if( c == ' ' )
    	*(d++) = '_';
    else
    if( ( ( c >= 'A' ) && ( c <= 'Z' ) ) ||
        ( ( c >= 'a' ) && ( c <= 'z' ) ) ||
        ( ( c >= '0' ) && ( c <= '9' ) ) )
        	*(d++) = c;
    else {
      sprintf( d,"=%02X",((int)c) & 0xFF );
      d += 3;
    }
  }
  *d = 0;
  strcat( dest,"?=" );
}

void utf8touni( unsigned char *buffer,unsigned int *unibuffer, int buflen ) {
  int ptr = 0;
  unsigned char leadchar;
  unsigned int resultchar;
  int pcctr;
  unsigned char pcmask;
  int	i;
  
  while( *buffer ) {
    leadchar = *buffer++;
    resultchar = 0;
    if( ( leadchar & 0x80 ) == 0 ) {
      pcctr = 0;
      pcmask = 0x7F;
    } else
    if( ( leadchar & 0xE0 ) == 0xC0 ) {
      pcctr = 1;
      pcmask = 0x1F;
    } else
    if( ( leadchar & 0xF0 ) == 0xE0 ) {
      pcctr = 2;
      pcmask = 0x0F;
    } else
    if( ( leadchar & 0xF8 ) == 0xF0 ) {
      pcctr = 3;
      pcmask = 0x07;
    } else
    if( ( leadchar = 0xFC ) == 0xF8 ) {
      pcctr = 4;
      pcmask = 0x03;
    } else
    if( ( leadchar = 0xFE ) == 0xFC ) {
      pcctr = 5;
      pcmask = 0x01;
    } else {
invalidchar:
	unibuffer[ptr++] = (unsigned int)'?';
nextchar:
	if( ptr >= buflen )
		break;
	else
		continue;
    }
    resultchar = (unsigned int)( leadchar & pcmask );
    for( i = 0 ; i < pcctr ; ++i ) {
      leadchar = *(buffer++);
      if( leadchar == 0 )
      	goto endloop;
      if( ( leadchar & 0xC0 ) != 0x80 )
      	goto invalidchar;
      resultchar <<= 6;
      resultchar |= (unsigned int)( leadchar & 0x3F);
    }
    unibuffer[ptr++] = resultchar;
    goto nextchar;
  }
endloop:
  unibuffer[ptr] = 0;
}

void unitoiso8859_1( unsigned int *unibuffer, char *tobuffer ) {
  unsigned int unichar;
  
  while( *unibuffer ) {
    unichar = *(unibuffer++);
    if( unichar < 0x100 )
    	*(tobuffer++) = (char)unichar;
    else
    	*(tobuffer++) = '?';
  }
  *tobuffer = 0;
}

/* Transfers a string coded in fromchset encoding to tochset encoding.
   *** WARNING *** Temporary implementation. This time it transfers only
   utf-8 to iso-8859-1. */
void transfer_charset( char frombuffer[], char tobuffer[], int fromchset, int tochset ) {
  unsigned int	unibuffer[256];
  if( ( fromchset == CHARSET_UTF_8 ) && 
      ( tochset == CHARSET_ISO_8859_1 ) ) {
      utf8touni( (unsigned char *)frombuffer,unibuffer,255 );
      unitoiso8859_1( unibuffer,tobuffer );
  } else
  	strcpy( tobuffer,frombuffer );
}
